#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <gnome-keyring.h>

#define  MINIMUM_UID 500

static char  *cache_data = NULL;
static size_t cache_used = 0;
static size_t cache_size = 0;
static char  *local_pass = NULL;
static char *remote_pass = NULL;

static void cleanup(void)
{
    if (cache_size > 0) {
        memset(cache_data, '\0', cache_size);
        munlock(cache_data, cache_size);
        munmap(cache_data, cache_size);
        cache_size = 0;
        cache_used = 0;
        cache_data = NULL;
    }
    local_pass = NULL;
    remote_pass = NULL;
}

static int get_passwords(const int descriptor)
{
    struct termios  oldterm, newterm;
    int             retval = 0;
    long            page;

    local_pass = NULL;
    remote_pass = NULL;
    cache_data = NULL;
    cache_size = 0;
    cache_used = 0;

    if (descriptor == -1)
        return errno = EINVAL;

    page = sysconf(_SC_PAGESIZE);
    if (page < 1L)
        return errno = ENOTSUP;

    /* Memory-map a single page for the passwords. */
    cache_size = page;
    cache_data = mmap(NULL, cache_size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, (off_t)0);
    if (cache_data == MAP_FAILED) {
        cache_size = 0;
        cache_used = 0;
        cache_data = NULL;
        return errno = ENOMEM;
    }

    /* Clear before locking. This should make sure the page is populated. */
    memset(cache_data, '\0', cache_size);

    if (mlock(cache_data, cache_size)) {
        munmap(cache_data, cache_size);
        cache_size = 0;
        cache_used = 0;
        cache_data = NULL;
        return errno = ENOMEM;
    }

    /* If the descriptor refers to a terminal, disable echo. */
    if (isatty(descriptor)) {
        if (tcgetattr(descriptor, &oldterm)) {
            retval = errno;
            munlock(cache_data, cache_size);
            munmap(cache_data, cache_size);
            cache_size = 0;
            cache_used = 0;
            cache_data = NULL;
            return errno = retval;
        }
        memcpy(&newterm, &oldterm, sizeof newterm);
        newterm.c_lflag &= ~ECHO;
        if (tcsetattr(descriptor, TCSANOW, &newterm)) {
            retval = errno;
            munlock(cache_data, cache_size);
            munmap(cache_data, cache_size);
            cache_size = 0;
            cache_used = 0;
            cache_data = NULL;
            return errno = retval;
        }
    }

    /* Clear after locking. This is probably paranoid.. */
    memset(cache_data, '\0', cache_size);

    /* Read until EOF, or cache full. */
    while (1) {
        ssize_t bytes;

        if (cache_used + 1 >= cache_size) {
            retval = ENOSPC;
            break;
        }

        bytes = read(descriptor, cache_data + cache_used, cache_size - cache_used - 1);
        if (bytes > (ssize_t)0)
            cache_used += bytes;
        else
        if (bytes == (ssize_t)0)
            break;
        else
        if (bytes != (ssize_t)-1) {
            retval = EIO;
            break;
        } else
        if (errno != EINTR) {
            retval = errno;
            break;
        }
    }

    /* If terminal, revert to previous settings. */
    if (isatty(descriptor)) {
        tcflush(descriptor, TCIOFLUSH);
        if (tcsetattr(descriptor, TCSANOW, &oldterm))
            if (!retval)
                retval = errno;
        tcflush(descriptor, TCIOFLUSH);
    }

    /* Parse the data. */
    do {
        size_t len1, off2, len2;

        /* Assume invalid data, so we can just break. */
        retval = ENODATA;

        /* No data? Bug? */
        if (cache_used < 2 || cache_used >= cache_size)
            break;

        /* Terminate cached data, so we can treat it as a string. */
        cache_data[cache_used] = '\0';

        /* Length of the first line. May be empty. */
        len1 = strcspn(cache_data, "\r\n");
        if (len1 >= cache_used)
            break;

        /* Offset to start of second line. */
        off2 = len1 + strspn(cache_data + len1, "\r\n");
        if (off2 <= len1 || off2 >= cache_used)
            break;

        /* Length of the second line. */
        len2 = strcspn(cache_data + off2, "\r\n");
        if (len2 < 1 || off2 + len2 >= cache_used)
            break;

        /* Make sure second line ends with a newline. */
        if (cache_data[off2 + len2] != '\r' &&
            cache_data[off2 + len2] != '\n')
            break;

        /* Separate lines into strings */
        cache_data[len1] = '\0';
        cache_data[off2 + len2] = '\0';

        /* Save pointers. */
        local_pass = cache_data;
        remote_pass = cache_data + off2;

        /* Parsed successfully. */
        retval = 0;
    } while (0);

    /* If an error occurred, clean up the cache and return. */
    if (retval) {
        memset(cache_data, '\0', cache_size);
        munlock(cache_data, cache_size);
        munmap(cache_data, cache_size);
        cache_data = NULL;
        cache_size = 0;
        cache_used = 0;
        local_pass = NULL;
        remote_pass = NULL;
        return errno = retval;
    }

    /* Done. */
    return errno = 0;
}

int main(int argc, char *argv[])
{
    GnomeKeyringAttributeList *attributes = NULL;
    GnomeKeyringResult result;
    guint32 id;
    int arg;

    if (argc < 2 || (argc % 2) == 1 || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--help")) {
        fprintf(stderr, "\n");
        fprintf(stderr, "Usage: %s [ -h | --help ]\n", argv[0]);
        fprintf(stderr, "       %s KEY [ ATTR VALUE ]*\n", argv[0]);
        fprintf(stderr, "\n");
        fprintf(stderr, "This utility updates the gnome-keyring key KEY\n");
        fprintf(stderr, "(of GNOME_KEYRING_ITEM_GENERIC_SECRET type)\n");
        fprintf(stderr, "belonging to the current user.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Standard input should contain the local/keyring password\n");
        fprintf(stderr, "on the first line, and the secret on the second line.\n");
        fprintf(stderr, "Echo is disabled, and no prompt is shown.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "For NetworkManager WiFi connections, you'll want to use\n");
        fprintf(stderr, "\t'Network secret for <ID>/802-1x/password'\n");
        fprintf(stderr, "\t'xdg:schema' 'org.freedesktop.NetworkManager.Connection'\n");
        fprintf(stderr, "\t'setting-key' 'password'\n");
        fprintf(stderr, "\t'setting-name' '802-1x'\n");
        fprintf(stderr, "\t'connection-uuid' '<UUID>'\n");
        fprintf(stderr, "where <ID> and <UUID> are the corresponding entries in\n");
        fprintf(stderr, "the NetworkManager connection file. The secret specified\n");
        fprintf(stderr, "in the second line of standard input is the WiFi password.\n");
        fprintf(stderr, "\n");
        return EXIT_FAILURE;
    }

    if (getuid() < (uid_t)MINIMUM_UID) {
        fprintf(stderr, "You are not allowed to run this utility (UID is too small).\n");
        return EXIT_FAILURE;
    }

    switch (get_passwords(STDIN_FILENO)) {
    case 0: break;

    case ENOMEM:
        fprintf(stderr, "Too much input.\n");
        return EXIT_FAILURE;

    case ENODATA:
        fprintf(stderr, "No passwords specified.\n");
        return EXIT_FAILURE;

    default:
        fprintf(stderr, "Error reading passwords from standard input.\n");
        return EXIT_FAILURE;
    }

    attributes = gnome_keyring_attribute_list_new();
    for (arg = 2; arg + 1 < argc; arg += 2)
        gnome_keyring_attribute_list_append_string(attributes, argv[arg+0], argv[arg+1]);

#ifdef CREATE_KEYRING

    /* Delete the "login" keyring. */
    result = gnome_keyring_delete_sync("login");
    if (result != GNOME_KEYRING_RESULT_OK &&
        result != GNOME_KEYRING_RESULT_NO_SUCH_KEYRING) {
        fprintf(stderr, "Cannot delete 'login' keyring: %s.\n", gnome_keyring_result_to_message(result));
        gnome_keyring_attribute_list_free(attributes);
        cleanup();
        return EXIT_FAILURE;
    }

    /* Create the "login" keyring. */
    result = gnome_keyring_create_sync("login", local_pass);
    if (result != GNOME_KEYRING_RESULT_OK) {
        fprintf(stderr, "Cannot create 'login' keyring: %s.\n", gnome_keyring_result_to_message(result));
        gnome_keyring_attribute_list_free(attributes);
        cleanup();
        return EXIT_FAILURE;
    }

    /* Unlock the keyring. */
    result = gnome_keyring_unlock_sync("login", local_pass);
    if (result != GNOME_KEYRING_RESULT_ALREADY_UNLOCKED &&
        result != GNOME_KEYRING_RESULT_OK) {
        fprintf(stderr, "Error unlocking newly-created 'login' keyring: %s.\n", gnome_keyring_result_to_message(result));
        gnome_keyring_attribute_list_free(attributes);
        cleanup();
        return EXIT_FAILURE;
    }

#else

    /* Unlock the keyring. */
    result = gnome_keyring_unlock_sync("login", local_pass);
    if (result != GNOME_KEYRING_RESULT_ALREADY_UNLOCKED &&
        result != GNOME_KEYRING_RESULT_OK) {
        fprintf(stderr, "Error unlocking 'login' keyring: %s.\n", gnome_keyring_result_to_message(result));
        gnome_keyring_attribute_list_free(attributes);
        cleanup();
        return EXIT_FAILURE;
    }

#endif

    /* Create the specified entry. */
    result = gnome_keyring_item_create_sync("login",
                                            GNOME_KEYRING_ITEM_GENERIC_SECRET,
                                            argv[1],
                                            attributes,
                                            remote_pass,
                                            FALSE,
                                            &id);
    if (result != GNOME_KEYRING_RESULT_OK) {
        fprintf(stderr, "Error creating the key: %s.\n", gnome_keyring_result_to_message(result));
        gnome_keyring_attribute_list_free(attributes);
        cleanup();
        return EXIT_FAILURE;
    }

    /* Cleanup. */
    gnome_keyring_attribute_list_free(attributes);
    cleanup();

    /* Success. */
    return EXIT_SUCCESS;
}
