#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <gnome-keyring.h>

static char *read_secret(const int descriptor)
{
    struct termios original;
    int            terminal = 0;
    int            failure = 0;
    int            result;

    char          *data = NULL;
    size_t         size = 0;
    size_t         used = 0;
    ssize_t        n;

    if (descriptor == -1) {
        errno = EINVAL;
        return NULL;
    }

    memset(&original,  0, sizeof original);
    result = tcgetattr(descriptor, &original);
    if (!result) {
        struct termios temporary;
        memcpy(&temporary, &original, sizeof temporary);
        temporary.c_lflag &= ~(ECHO | ECHONL);
        if (!tcsetattr(STDIN_FILENO, TCSANOW, &temporary))
            terminal = 1;
    }

    while (1) {

        if (used >= size) {
            char *temp;
            size = (used | 511) + 1;
            temp = realloc(data, size);
            if (!temp) {
                failure = ENOMEM;
                break;
            }
            data = temp;
        }

        n = read(descriptor, data + used, 1);
        if (n == (ssize_t)1) {
            if (data[used] == '\n') {
                data[used] = '\0';
                used++;
                break;
            } else
            if ((data[used] >= 0 && data[used] < 32) || (data[used] == 127)) {
                data[used] = '\0';
                failure = EDOM;
                break;
            } else
                used++;
        } else
        if (n == (ssize_t)0) {
            data[used] = '\0';
            break;
        } else
        if (n != (ssize_t)-1) {
            failure = EIO;
            break;
        } else
        if (errno != EINTR && errno) {
            failure = errno;
            break;
        }
    }

    if (terminal)
        tcsetattr(STDIN_FILENO, TCSANOW, &original);

    if (failure) {
        free(data);
        errno = failure;
        return NULL;
    }

    errno = 0;
    return data;
}

static char *free_secret(char *const ptr)
{
    if (ptr) {
        const size_t len = strlen(ptr);
        if (len)
            memset(ptr, 0, len);
        free(ptr);
    }
    return NULL;
}

int main(int argc, char *argv[])
{
    const char         *keyring, *keyname;
    char               *secret;
    GnomeKeyringResult  result;
    int                 arg;

    if (argc < 2 || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--help")) {
        fprintf(stderr, "\n");
        fprintf(stderr, "Usage: %s [ -h | --help ]\n", argv[0]);
        fprintf(stderr, "       %s keyring [ [ -g keyname ] -a name value ]*\n", argv[0]);
        fprintf(stderr, "\n");
        fprintf(stderr, "Initial keyring password is supplied in standard input.\n");
        fprintf(stderr, "The secret corresponding to each key (-g) is also supplied\n");
        fprintf(stderr, "in standard input.\n");
        fprintf(stderr, "\n");
        return EXIT_FAILURE;
    }

    if (!gnome_keyring_is_available()) {
        fprintf(stderr, "Keyring service is not available.\n");
        return EXIT_FAILURE;
    }

    keyring = argv[1];
    arg = 2;
    secret = read_secret(STDIN_FILENO);
    if (!secret) {
        fprintf(stderr, "No password specified for keyring '%s'.\n", keyring);
        return EXIT_FAILURE;
    }

    /* Delete any existing keyring. */
    result = gnome_keyring_delete_sync(keyring);
    if (result != GNOME_KEYRING_RESULT_OK &&
        result != GNOME_KEYRING_RESULT_NO_SUCH_KEYRING) {
        fprintf(stderr, "Cannot delete '%s' keyring: %s.\n", keyring, gnome_keyring_result_to_message(result));
        return EXIT_FAILURE;
    }

    /* Create and unlock the "login" keyring. */
    result = gnome_keyring_create_sync(keyring, secret);
    if (result != GNOME_KEYRING_RESULT_OK) {
        fprintf(stderr, "Cannot create '%s' keyring: %s.\n", keyring, gnome_keyring_result_to_message(result));
        return EXIT_FAILURE;
    }
    result = gnome_keyring_unlock_sync(keyring, secret);
    if (result != GNOME_KEYRING_RESULT_ALREADY_UNLOCKED &&
        result != GNOME_KEYRING_RESULT_OK) {
        fprintf(stderr, "Cannot unlock '%s' keyring: %s.\n", keyring, gnome_keyring_result_to_message(result));
        return EXIT_FAILURE;
    }
    free_secret(secret);
    secret = NULL;

    printf("Keyring '%s' created successfully.\n", keyring);
    fflush(stdout);

    /* Add new keys. */
    while (arg < argc) {
        GnomeKeyringAttributeList *attrlist = NULL;
        GnomeKeyringItemType       type;
        guint32                    id = 0;

        if (!strcmp(argv[arg], "-g"))
            type = GNOME_KEYRING_ITEM_GENERIC_SECRET;
        else
        if (!strcmp(argv[arg], "-p"))
            type = GNOME_KEYRING_ITEM_NETWORK_PASSWORD;
        else {
            fprintf(stderr, "%s: Unsupported option.\n", argv[arg]);
            return EXIT_FAILURE;
        }

        if (++arg >= argc) {
            fprintf(stderr, "Missing key name for keyring '%s'.\n", keyring);
            return EXIT_FAILURE;
        }

        keyname = argv[arg++];
        if (!keyname || !*keyname) {
            fprintf(stderr, "Empty key name for keyring '%s'.\n", keyring);
            return EXIT_FAILURE;
        }

        secret = read_secret(STDIN_FILENO);
        if (!secret) {
            fprintf(stderr, "No secret given for key '%s'.\n", keyname);
            return EXIT_FAILURE;
        }

        attrlist = gnome_keyring_attribute_list_new();
        while (arg + 2 < argc && !strcmp(argv[arg], "-a")) {
           gnome_keyring_attribute_list_append_string(attrlist, argv[arg + 1], argv[arg + 2]);
           arg += 3;
        }

        result = gnome_keyring_item_create_sync(keyring, type, keyname, attrlist, secret, FALSE, &id);
        gnome_keyring_attribute_list_free(attrlist);
        free_secret(secret);
        attrlist = NULL;
        secret = NULL;
        if (result != GNOME_KEYRING_RESULT_OK) {
            fprintf(stderr, "Cannot add key '%s' to keyring '%s': %s.\n", keyname, keyring, gnome_keyring_result_to_message(result));
            return EXIT_FAILURE;
        }

        printf("Key '%s' added to keyring '%s' successfully (id %u).\n", keyname, keyring, (unsigned int)id);
        fflush(stdout);
    }

    return EXIT_SUCCESS;
}
