#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "config.h"

static const char dir[] = ROOTSCRIPTS "/";

const char *executable(const char *const name)
{
    const uid_t          real_uid = getuid();
    const gid_t          real_gid = getgid();
    const struct passwd *real_user;
    const struct group  *real_group;
    size_t               dirlen, namelen, size;
    char                *newpath;

    /* If no script named, we cannot find it. */
    if (!name || !*name) {
        errno = ENOENT;
        return NULL;
    }

    /* Verify name is valid. */
    namelen = strlen(name);
    if (namelen < 1) {
        errno = ENOENT;
        return NULL;
    }
    if (strspn(name, "0123456789-abcdefghijklmnopqrstuvwxyz") != namelen) {
        errno = ENOENT;
        return NULL;
    }
    
    /* Make sure we are run by an user (not root or system account). */
    if (real_uid < (uid_t)MIN_UID ||
        real_gid < (uid_t)MIN_GID) {
        errno = EACCES;
        return NULL;
    }

    /* Find out the only user we accept. */
    real_group = getgrnam(GROUP);
    if (!real_group || real_group->gr_gid != real_gid) {
        errno = EACCES;
        return NULL;
    }
    real_user = getpwnam(USER);
    if (!real_user || real_user->pw_uid != real_uid ||
        real_user->pw_gid != real_gid) {
        errno = EACCES;
        return NULL;
    }

    /* Path prefix. */
    dirlen = strlen(dir);
    if (dirlen < 1 || dir[0] != '/') {
        errno = ENOMEM;
        return NULL;
    }
    size = ((dirlen + namelen) | 15) + 17;
    newpath = malloc(size);
    if (!newpath) {
        errno = ENOMEM;
        return NULL;
    }
    memset(newpath, '\0', size);
    memcpy(newpath, dir, dirlen);
    memcpy(newpath + dirlen, name, namelen);

    /* exec() will take care of permissions checks. */
    errno = 0;
    return newpath;
}    

int main(int argc, char *argv[])
{
    uid_t ruid, euid, suid;
    gid_t rgid, egid, sgid;
    const char *path;

    if (getresuid(&ruid, &euid, &suid) ||
        getresgid(&rgid, &egid, &sgid)) {
        fprintf(stderr, "Cannot get identity: %m.\n");
        return EXIT_FAILURE;
    }

    if (euid != (uid_t)0) {
        fprintf(stderr, "Not installed as setuid root.\n");
        return EXIT_FAILURE;
    }

    if (argc < 2 || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--help")) {
        fprintf(stderr, "\n");
        fprintf(stderr, "Usage: %s [ -h | --help ]\n", argv[0]);
        fprintf(stderr, "       %s script [ arguments ... ]\n", argv[0]);
        fprintf(stderr, "\n");
        fprintf(stderr, "Execute a script in %s with root privileges.\n", dir);
        fprintf(stderr, "\n");
        return EXIT_FAILURE;
    }

    path = executable(argv[1]);
    if (!path) {
        fprintf(stderr, "%s: %m.\n", argv[1]);
        return EXIT_FAILURE;
    }

    if (setresgid((gid_t)0, (gid_t)0, (gid_t)0)) {
        fprintf(stderr, "Cannot set group identity: %m.\n");
        return EXIT_FAILURE;
    }
    if (setresuid((uid_t)0, (uid_t)0, (uid_t)0)) {
        fprintf(stderr, "Cannot set user identity: %m.\n");
        return EXIT_FAILURE;
    }

    execv(path, argv + 1);
    fprintf(stderr, "%s: %m.\n", argv[1]);
    return EXIT_FAILURE;
}
