#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "config.h"

static const char dir[] = USERSCRIPTS "/";

const char **read_environment(const char *const filename)
{
    char   *data = NULL;
    size_t  size = 0;
    size_t  used = 0;

    char  **pair = NULL;
    size_t  pairs = 0;
    size_t  pairs_max = 0;

    int     fd;
    ssize_t n;
    if (!filename || !*filename) {
        errno = EINVAL;
        return NULL;
    }

    fd = open(filename, O_RDONLY | O_NOCTTY);
    if (fd == -1)
        return NULL;

    while (1) {
        if (used >= size) {
            const size_t new_size = (used | 4095) + 4095;
            char        *new_data;

            new_data = realloc(data, new_size + 2);
            if (!new_data) {
                close(fd);
                free(data);
                errno = ENOMEM;
                return NULL;
            }

            size = new_size;
            data = new_data;
        }

        n = read(fd, data + used, size - used);
        if (n > (ssize_t)0)
            used += n;
        else
        if (n == (ssize_t)0)
            break;
        else
        if (n != (ssize_t)-1 || errno != EINTR) {
            close(fd);
            free(data);
            errno = EIO;
            return NULL;
        }
    }

    if (!used) {
        free(data);
        errno = ENODATA;
        return NULL;
    }

    if (close(fd)) {
        free(data);
        errno = EIO;
        return NULL;
    }

    {
        char       *p = data;
        char *const q = data + used;

        while (p < q)
            if (*p == '\n')
                *(p++) = '\0';
            else
                p++;

        p = data;

        while (1) {

            while (p < q && !*p)
                p++;
            if (p >= q)
                break;

            if (pairs >= pairs_max) {
               char  **new_pair;
                size_t  new_max = (pairs | 255) + 256;

                new_pair = realloc(pair, (new_max + 1) * sizeof (char *));
                if (!new_pair) {
                    free(pair);
                    free(data);
                    errno = ENOMEM;
                    return NULL;
                }

                pair = new_pair;
                pairs_max = new_max;
            }

            pair[pairs++] = p;

            while (p < q && *p != '\0')
                p++;
            *(p++) = '\0';
        }
    }

    if (pairs < 1) {
        free(pair);
        free(data);
        errno = ENODATA;
        return NULL;
    }

    pair[pairs] = NULL;
    return (const char **)pair;
}

const char *executable(const char *const name)
{
    size_t  dirlen, namelen, size;
    char   *newpath;

    /* If no script named, we cannot find it. */
    if (!name || !*name) {
        errno = ENOENT;
        return NULL;
    }

    /* Verify name is valid. */
    namelen = strlen(name);
    if (namelen < 1) {
        errno = ENOENT;
        return NULL;
    }
    if (strspn(name, "0123456789-abcdefghijklmnopqrstuvwxyz") != namelen) {
        errno = ENOENT;
        return NULL;
    }

    /* Path prefix. */
    dirlen = strlen(dir);
    if (dirlen < 1 || dir[0] != '/') {
        errno = ENOMEM;
        return NULL;
    }
    size = ((dirlen + namelen) | 15) + 17;
    newpath = malloc(size);
    if (!newpath) {
        errno = ENOMEM;
        return NULL;
    }
    memset(newpath, '\0', size);
    memcpy(newpath, dir, dirlen);
    memcpy(newpath + dirlen, name, namelen);

    /* exec() will take care of permissions checks. */
    errno = 0;
    return newpath;
}


int main(int argc, char *argv[])
{
    uid_t ruid, euid, suid, as_uid;
    gid_t rgid, egid, sgid, as_gid;
    const char *user, *envfile;
    const char *path, **env;

    if (getresuid(&ruid, &euid, &suid) ||
        getresgid(&rgid, &egid, &sgid)) {
        fprintf(stderr, "Cannot get identity: %m.\n");
        return EXIT_FAILURE;
    }

    if (euid != (uid_t)0) {
        fprintf(stderr, "Not installed as setuid root.\n");
        return EXIT_FAILURE;
    }

    if (argc < 4 || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--help")) {
        fprintf(stderr, "\n");
        fprintf(stderr, "Usage: %s [ -h | --help ]\n", argv[0]);
        fprintf(stderr, "       %s user envfile script [ arguments ... ]\n", argv[0]);
        fprintf(stderr, "\n");
        fprintf(stderr, "Execute a script in %s with user privileges.\n", dir);
        fprintf(stderr, "\n");
        return EXIT_FAILURE;
    }

    argv[argc] = NULL;
    user = argv[1];
    envfile = argv[2];

    if (!user || !*user) {
        fprintf(stderr, "No user specified.\n");
        return EXIT_FAILURE;

    } else {
        const struct passwd *const as = getpwnam(user);

        if (!as) {
            fprintf(stderr, "%s: %m.\n", argv[1]);
            return EXIT_FAILURE;
        }

        as_uid = as->pw_uid;
        as_gid = as->pw_gid;
        if (as_uid < MIN_UID || as_gid < MIN_GID) {
            fprintf(stderr, "%s: Invalid user.\n", argv[1]);
            return EXIT_FAILURE;
        }
    }

    env = read_environment(envfile);
    if (!env) {
        fprintf(stderr, "%s: Cannot read environment: %m.\n", envfile);
        return EXIT_FAILURE;
    }

    path = executable(argv[3]);
    if (!path) {
        fprintf(stderr, "%s: %m.\n", argv[1]);
        return EXIT_FAILURE;
    }

    if (setresgid((gid_t)as_gid, (gid_t)as_gid, (gid_t)as_gid)) {
        fprintf(stderr, "Cannot set group identity: %m.\n");
        return EXIT_FAILURE;
    }
    if (initgroups(argv[1], as_gid)) {
        fprintf(stderr, "Cannot set supplementary group identities: %m.\n");
        return EXIT_FAILURE;
    }

    if (setresuid((uid_t)as_uid, (uid_t)as_uid, (uid_t)as_uid)) {
        fprintf(stderr, "Cannot set user identity: %m.\n");
        return EXIT_FAILURE;
    }

    execve(path, argv + 3, env);
    fprintf(stderr, "%s: %m.\n", argv[3]);
    return EXIT_FAILURE;
}
